#!/usr/bin/env python3

import argparse
import base64
import json
import hashlib
import logging
import os
import requests
import time
import urllib3

from kubernetes import client, config

import rbac_util_encoding

urllib3.disable_warnings()
logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)s: %(message)s')


def get_claims(org_key, target_connector_id, auth_connector_id, api_key, env=None):
    uri = f"{CBC_HOST}/access/v2/orgs/{org_key}/connectors/{target_connector_id}/claims?type=CUSTOM"
    resp = requests.get(uri, headers={ "X-Auth-Token": f"{api_key}/{auth_connector_id}"})
    claims = ""
    if resp.status_code == 200:
        try:
            claims = json.loads(resp.text)["claims"]
        except:
            logging.error(f"Failed to get claims via '{uri}'")
            return False
    else:
        logging.error(f"Request to '{uri}' returned: '{resp.text}'")
        return False

    if claims is not "null":
        logging.info(f"Claims returned: '{claims}'")
        return True
    else:
        logging.error(f"Claims returned 'null'")
        return False



IS_K8S = (os.getenv("IS_K8S", 'False') == 'true')
if not IS_K8S:
    print("No further actions, for this is not a K8S environment.")
    exit(0)

RELEASE_VERSION = os.getenv("RELEASE_VERSION")
ENV = os.getenv("ENV")
MCS2_ENV = os.getenv("MCS2_ENV")
CBC_HOST = os.getenv("CBC_HOST")
MOSAIC_APPS_NAMESPACE = os.getenv("NS_APPS", f"cb-{MCS2_ENV}-apps")
MOSAIC_INFRA_NAMESPACE = os.getenv("NS_INFRA", f"cb-{MCS2_ENV}-infra")
MOSAIC_NAMESPACES = [ MOSAIC_APPS_NAMESPACE, MOSAIC_INFRA_NAMESPACE ]

connectors_directory = os.path.dirname(os.path.abspath(__file__ + "/.."))

parser = argparse.ArgumentParser(description="Generate service connectors and associate grants")
parser.add_argument("--salt", help="The salt use to encode connectorId and apiKey. If omitted, the org-key will be used as a salt.", 
                              dest='salt', 
                              required=False)

args = parser.parse_known_args()[0]
org_key = rbac_util_encoding.encode_org_key(env=ENV, org_id=1) 
salt = org_key
if args.salt:
    salt = salt + args.salt

## Kubernetes prep
# Load the cluster config, which is handled by the library, automatically.
# It is assumed that there is a ServiceAccount attached to the pod, executing this script.
config.load_incluster_config()
core_api_v1 = client.CoreV1Api()
# Assure scmigration job has finished succesfully
retries = 0
max_retries = 20
initial_delay = 10
max_delay = 60
job_name = "platform-scmigration-job"
label_selector = f"job-name={job_name},app.kubernetes.io/version={RELEASE_VERSION}"
break_flag = False
while retries <= max_retries:
    if retries == max_retries:
        raise Exception(f"Timeout waiting for '{job_name}'.")
    
    pods = core_api_v1.list_namespaced_pod(namespace=MOSAIC_APPS_NAMESPACE,
                                           label_selector=label_selector).items

    if pods:
        for pod in pods:
            pod_name = pod.metadata.name
            pod_status = str(pod.status.phase)
            logging.info(f"Pod '{pod_name}' associated with Job '{job_name}' has status: {pod_status}")
            if pod_status == "Succeeded":
                break_flag = True
            elif pod_status == "Failed":
                raise Exception("'{job_name}' has failed its execution. Connector secret changes will not be executed.")
    
    if break_flag:
        break
    
    logging.info(f"Waiting for Job '{job_name}'. Status check in {initial_delay} seconds...")
    time.sleep(initial_delay)
    initial_delay = min(2 * initial_delay, max_delay)
    retries += 1

# Grant is guaranteed to exist now; confirm this via claims requests
# Promote the -temp connector Kubernetes secret
logging.info("Starting connector claims checks and secret migration...")
logging.info("-----------------------------------------------------------------")
for filename in os.listdir("{}/connectors/orgs/1".format(connectors_directory)):
    if filename.endswith(".json"):
        try:
            f = open("{}/connectors/orgs/1/{}".format(connectors_directory, filename), 'r')
        except IOError:
            logging.error("An IO error occured when reading from file roles/{}".format(filename))
        else:
            text = f.read()
            service_connector_json = json.loads(text)
            service_connector_name = service_connector_json['connector_name']

            # Actual length is dklen*2
            connector_id = rbac_util_encoding.encode_connector_id(service_connector_name, salt)
            api_key = rbac_util_encoding.encode_api_key(service_connector_name, salt)

            connector_secret_name = service_connector_name.replace('_', '-').lower()
            temp_secret_name = f"{connector_secret_name}-temp"
            claims_tested = False
            for namespace in MOSAIC_NAMESPACES:
                try:
                    temp_secret = core_api_v1.read_namespaced_secret(temp_secret_name, namespace).data
                except client.rest.ApiException:
                    logging.info(f"No temp secret for '{connector_secret_name}' in '{namespace}'.")
                    original_secret = core_api_v1.read_namespaced_secret(connector_secret_name, namespace).data
                    if not claims_tested:
                        logging.info(f"Get claims using credentials from existing secret.")
                        if 'multienv_connector_json' in original_secret:
                            multienv_json = json.loads(base64.b64decode(original_secret['multienv_connector_json']).decode('utf-8'))
                            for env in multienv_json:
                                env_orgkey = rbac_util_encoding.encode_org_key(env=env, org_id=1)
                                # ENV is the master environment, only its credentials are used
                                if not get_claims(org_key=env_orgkey,
                                                  target_connector_id=multienv_json[env]['connector-id'],
                                                  auth_connector_id=multienv_json[ENV]['connector-id'], 
                                                  api_key=multienv_json[ENV]['api-key']):
                                    raise Exception("Failed claims check with credentials from secret '{connector_secret_name}' for environment '{env}'")
                            claims_tested = True
                        else:
                            if get_claims(org_key=org_key, 
                                          target_connector_id=base64.b64decode(original_secret['connector-id']).decode('utf-8'), 
                                          auth_connector_id=base64.b64decode(original_secret['connector-id']).decode('utf-8'),
                                          api_key=base64.b64decode(original_secret['api-key']).decode('utf-8')):
                                claims_tested = True
                    continue
    
                # Double check connector-id and api-key and make a claims call to check validity
                if connector_id != base64.b64decode(temp_secret['connector-id']).decode('utf-8') or api_key != base64.b64decode(temp_secret['api-key']).decode('utf-8'):
                    raise Exception(f"Mismatch between scmigration generated and live-migration generated credentials in '{temp_secret_name}'")
                
                if not claims_tested:
                    logging.info(f"Get claims with '{temp_secret_name}'")
                    if get_claims(org_key=org_key, target_connector_id=connector_id, auth_connector_id=connector_id, api_key=api_key):
                        claims_tested = True
                    else:
                        raise Exception("Failed claims check with credentials from secret '{temp_secret_name}'")
    
                    # Check if MULTIENV secret
                    if 'multienv_connector_json' in temp_secret:
                        multienv_json = json.loads(base64.b64decode(temp_secret['multienv_connector_json']).decode('utf-8'))
                        for env in multienv_json:
                            ex_connector_id = multienv_json[env]['connector-id']
                            ex_api_key = multienv_json[env]['api-key']
                            env_orgkey = rbac_util_encoding.encode_org_key(env=env, org_id=1)
                            expected_connector_id = rbac_util_encoding.encode_connector_id(service_connector_name, env_orgkey)
                            expected_api_key = rbac_util_encoding.encode_api_key(service_connector_name, env_orgkey)
                            if ex_connector_id != expected_connector_id or ex_api_key != expected_api_key:
                                raise Exception(f"Mismatch between scmigration generated and live-migration generated credentials in '{temp_secret_name}' for environment {env}.")
                            logging.info(f"Get claims with '{temp_secret_name}' for specific environment '{env}'")
                            # ENV is the master environment, only its credentials are used
                            if get_claims(org_key=env_orgkey,
                                          target_connector_id=ex_connector_id,
                                          auth_connector_id=multienv_json[ENV]['connector-id'], 
                                          api_key=multienv_json[ENV]['api-key']):
                                claims_tested = True
                            else:
                                raise Exception("Failed claims check with credentials from secret '{temp_secret_name}' for environment '{env}'")
    
                # Replace the original secret
                new_secret = client.V1Secret(metadata=client.V1ObjectMeta(name=connector_secret_name,
                                                                          labels={"type": "connector"}),
                                             data=temp_secret)
                logging.info(f"Replacing '{connector_secret_name}' in '{namespace}'")
                core_api_v1.replace_namespaced_secret(name=connector_secret_name,
                                                      namespace=namespace,
                                                      body=new_secret)
                # Delete temp connector secret
                logging.info(f"Deleting '{temp_secret_name}' in '{namespace}'")
                core_api_v1.delete_namespaced_secret(temp_secret_name, namespace)
            logging.info("-----------------------------------------------------------------")

logging.info("All connector secrets have successfully been migrated.")
