import argparse
import base64
import yaml

from plans import *
from requests_sessions import *
from migration_executor import Executor

import os

def main(environment=None, x_auth_token=None, org_key=None, all=None, green_only=False, plan=False, csp_struct=None):
    script_file_path = os.path.dirname(os.path.abspath(__file__))
    entities_directory_path = os.path.dirname(os.path.abspath(__file__ + "/.."))
    environment_choices = ['localdev', 'dev08', 'dev01', 'test03', 'eap01', 'prod01', 'prod02', 'prod04',
                           'prod05', 'prod06', 'prodnrt', 'prodsyd']

    # Parse arguments
    parser = argparse.ArgumentParser(description="Migrate roles and permissions from change control.")
    parser.add_argument("--plan", help="Only output a summary of the expected changes the script would make",
                        action="store_true", default=plan)
    all_arg = parser.add_argument("--all",
                                  default=all,
                                  help="Migrate all roles and permissions from change control",
                                  action="store_true")
    parser.add_argument('-e', '--environment', dest='env', required=False, default=environment,
                        choices=environment_choices)
    parser.add_argument('-p', '--permissions', nargs='+', dest='permission_list', required=False)
    parser.add_argument('-r', '--roles', nargs='+', dest='role_list', required=False)
    parser.add_argument('-s', '--subroles', nargs='+', dest='subrole_list', required=False)
    parser.add_argument("-g", "--green-only",
                        default=green_only,
                        help="Run migration hitting only green service",
                        action="store_true")
    parser.add_argument('-u', "--url", help="Specifiy the base URL to access psc.access instead of using environment lookup tables",
                        dest='base_url', required=False)
    parser.add_argument("--request-context", help="Set the X-Request-Context header to this value - default omitted",
                        dest='request_context', required=False)
    parser.add_argument("--user-context", help="Set the X-User-Context header to this value - default omitted",
                        dest='user_context', required=False)
    parser.add_argument("--org-key", help="Set the org key for org 1",
                        dest='org_key', required=False)
    parser.add_argument("--csp-enabled", help="Flag that determines if CSP entities will be created",
                        dest='csp_enabled', required=False, action='store_true', default=False)
    parser.add_argument("--csp-url", help="URL to the given environment's CSP instance",
                        dest='csp_url', required=False)
    parser.add_argument("--csp-client-id", help="OAuth app ID",
                        dest='csp_client_id', required=False)
    parser.add_argument("--csp-client-secret", help="OAuth app secret",
                        dest='csp_client_secret', required=False)
    parser.add_argument("--definition-id", help="CSP service definition ID",
                        dest='definition_id', required=False)

    opts = parser.parse_known_args()[0]

    # validate environment and filter args
    validate_filter_arguments(opts, all_arg)
    if opts.base_url:
        base_url = opts.base_url
    else:
        env_urls = yaml.load(open('{}/url_configs.yml'.format(script_file_path)), Loader=yaml.BaseLoader)[opts.env]
        base_url = env_urls['dashboard']

    # Set filter list to be empty if the all flag was not set and no entities were provided of the type.
    # Only entities in the list will be planned / modified.  If the --all flag is set, set all filter lists
    # to None to prevent any filtering and modify all entities
    if not opts.all:
        if not opts.role_list:
            opts.role_list = []
        if not opts.permission_list:
            opts.permission_list = []
        if not opts.subrole_list:
            opts.subrole_list = []
    else:
        opts.role_list = None
        opts.permission_list = None
        opts.subrole_list = None

    if opts.org_key:
        org_key = opts.org_key
    if opts.env == 'localpipeline' or opts.env == 'localdev':
        org_key = 'R76KWHRN'

    # Build a common session object which store auth headers and base URL
    token = 'PZSGFNGMISYDXXXXXXXXX/ABCDEF1234' if x_auth_token is None else x_auth_token
    headers = {
        'X-Auth-Token': token,
        'Content-Type': 'application/json'
    }
    if opts.request_context:
        headers["X-Request-Context"] = opts.request_context
    if opts.user_context:
        headers["X-User-Context"] = opts.user_context

    session = MigrationSession(base_url, opts.green_only)
    session.headers.update(headers)

    # Build same session object, but for the CSP API
    csp_session = None
    if opts.csp_enabled:
        csp_credentials = base64.b64encode("{}:{}".format(opts.csp_client_id, opts.csp_client_secret).encode('utf-8')).decode('utf-8')
        csp_session = CSPSession(base_url=opts.csp_url, 
                                 oauth_credentials=csp_credentials, 
                                 service_definition_id=opts.definition_id)
        csp_session.headers.update({ "Content-Type": "application/json" })
    elif csp_struct:
        csp_credentials = base64.b64encode("{}:{}".format(csp_struct["client_id"], csp_struct["secret_id"]).encode('utf-8')).decode('utf-8')
        csp_session = CSPSession(base_url=csp_struct["url"],
                                 oauth_credentials=csp_credentials,
                                 service_definition_id=csp_struct["service_definition_id"])
        csp_session.headers.update({ "Content-Type": "application/json" })


    migration_plan = MigrationPlan(entities_directory_path, session, csp_session, org_key, opts.permission_list,
                                   opts.subrole_list, opts.role_list)

    migration_plan.print_plan()
    if not opts.plan:
        executor = Executor(org_key, session, csp_session, migration_plan)
        executor.execute_all()


def validate_filter_arguments(opts, all_arg):
    if not opts.role_list and not opts.permission_list and not opts.subrole_list and not opts.all:
        raise argparse.ArgumentError(all_arg, "--all Argument must be set if entities are not specified via args")
    elif (opts.role_list or opts.permission_list or opts.subrole_list) and opts.all:
        raise argparse.ArgumentError(all_arg, "--all Argument cannot be set if entities are specified via args")


if __name__ == "__main__":
    main()
