from enum import Enum
import os
import json
import requests
import time

class MigrationAction(Enum):
    """Enum indicating which action to take on a resource"""
    NOOP = 0
    CREATE = 1
    UPDATE = 2
    DELETE = 3

    def symbolize(self):
        if self.value is self.NOOP.value:
            return ""
        elif self.value is self.CREATE.value:
            return "C"
        elif self.value is self.UPDATE.value:
            return "U"
        elif self.value is self.DELETE.value:
            return "D"


class ResourcePlan():
    """Represents the planned actions to be taken on resources and their relationships"""

    def __init__(self, entity_json=None, entity_action="NOOP",
                 permissions_action="NOOP", subroles_action="NOOP",
                 overrides_action="NOOP"):
        self.entity_json = entity_json
        self.entity_action = entity_action
        self.permissions_action = permissions_action
        self.subroles_action = subroles_action
        self.overrides_action = overrides_action

    def print_plan(self):
        raise NotImplementedError("Not implemented")

    def clean_entity(self):
        raise NotImplementedError("Not implemented")


class SubrolePlan(ResourcePlan):
    def __init__(self, entity_json=None, entity_action="NOOP",
                 permissions_action="NOOP"):
        super().__init__(entity_json, entity_action, permissions_action)
        self.clean_entity()

    def print_plan(self):
        subrole_urn = self.entity_json['urn']
        def_symbol = self.entity_action
        perm_symbol = self.permissions_action
        print('%-75s|%-12s|%-12s' % (subrole_urn, def_symbol, perm_symbol))

    def clean_entity(self):
        """Remove the version fields from a subrole dict"""
        if 'version' in self.entity_json:
            self.entity_json.pop('version')
        if 'permissions' in self.entity_json:
            for perm in self.entity_json['permissions']:
                if 'version' in perm:
                    perm.pop('version')



class RolePlan(ResourcePlan):
    def __init__(self, entity_json=None, entity_action="NOOP",
                 permissions_action="NOOP", subroles_action="NOOP",
                 overrides_action="NOOP"):
        super().__init__(entity_json, entity_action, permissions_action, subroles_action, overrides_action)
        self.clean_entity()

    def print_plan(self):
        role_urn = self.entity_json['role_definition']['urn']
        def_symbol = self.entity_action
        perm_symbol = self.permissions_action
        sub_symbol = self.subroles_action
        overrides_symbol = self.overrides_action
        print('%-75s|%-12s|%-12s|%-12s|%-12s' % (role_urn, def_symbol, perm_symbol, sub_symbol, overrides_symbol))

    def clean_entity(self):
        """Remove the version fields from a role dict"""
        if 'version' in self.entity_json['role_definition']:
            self.entity_json['role_definition'].pop('version')
        if 'permissions' in self.entity_json and self.entity_json['permissions']:
            for perm in self.entity_json['permissions']:
                if 'version' in perm:
                    perm.pop('version')
        if 'subroles' in self.entity_json and self.entity_json['subroles']:
            for subrole in self.entity_json['subroles']:
                if 'version' in subrole:
                    subrole.pop('version')
        if 'overrides' in self.entity_json and self.entity_json['overrides']:
            for override in self.entity_json['overrides']:
                if 'version' in override:
                    override.pop('version')


class PermissionPlan(ResourcePlan):
    def __init__(self, entity_json=None, entity_action="NOOP"):
        super().__init__(entity_json, entity_action)

    def print_plan(self):
        permission_urn = self.entity_json['urn']
        def_symbol = self.entity_action
        print('%-75s|%-12s' % (permission_urn, def_symbol))

    def clean_entity(self):
        """Remove the version field from a permission dict"""
        if 'version' in self.entity_json:
            self.entity_json.pop('version')
        if 'overrides' in self.entity_json and not self.entity_json['overrides']:
            self.entity_json['overrides'] = []
        return self.entity_json


class MigrationPlan:
    """Represents the full migration plan"""

    def __init__(self, file_path, session, csp_session, org_key, permission_filter_list,
                 subroles_filter_list, role_filter_list):
        # Set context
        self.file_path = file_path
        self.session = session
        self.csp_session = csp_session
        self.org_key = org_key

        # Filter lists
        self.permission_filter_list = permission_filter_list
        self.subroles_filter_list = subroles_filter_list
        self.role_filter_list = role_filter_list

        # Build all resource plans
        self.permission_plan = self.plan_permission_migration(permission_filter_list)
        self.permission_plan_csp = None
        self.subrole_plan_csp = None
        if self.csp_session:
            self.permission_plan_csp = self.plan_permission_migration_csp(permission_filter_list)
            self.subrole_plan_csp = self.plan_subrole_migration_csp(subroles_filter_list)
        self.permission_plan = self.plan_permission_migration(permission_filter_list)
        self.subrole_plan = self.plan_subrole_migration(subroles_filter_list)
        self.role_plan = self.plan_role_migration(role_filter_list)
        self.org_role_plan = self.plan_role_migration(role_filter_list, is_org=True)

    def print_plan(self):
        print('Roles:')
        print('%-75s%-12s%-12s%-12s%-12s' % ("Role URN", "Definition", "Permissions", "subroles",
                                             "Overrides"))
        print('-' * 123)
        for resource_plan in self.role_plan:
            resource_plan.print_plan()

        print('\n\nOrg Roles:')
        print('%-75s%-12s%-12s%-12s%-12s' % ("Role URN", "Definition", "Permissions", "subroles",
                                             "Overrides"))
        print('-' * 123)
        for resource_plan in self.org_role_plan:
            resource_plan.print_plan()

        print('\n\nSubroles:')
        print('%-75s%-12s%-12s' % ("Subrole URN", "Definition", "Permissions"))
        print('-' * 123)
        for resource_plan in self.subrole_plan:
            resource_plan.print_plan()

        if self.subrole_plan_csp:
            print('\n\nCSP Subroles:')
            print('%-75s%-12s' % ("Subrole URN", "Definition"))
            print('-' * 123)
            for resource_plan in self.subrole_plan_csp:
                resource_plan.print_plan()        

        print('\n\nPermissions:')
        print('%-75s%-12s' % ("Permission URN", "Definition"))
        print('-' * 123)
        for resource_plan in self.permission_plan:
            resource_plan.print_plan()

        if self.permission_plan_csp:
            print('\n\nCSP Permissions:')
            print('%-75s%-12s' % ("Permission URN", "Definition"))
            print('-' * 123)
            for resource_plan in self.permission_plan_csp:
                resource_plan.print_plan()


    def plan_role_migration(self, filter_list, is_org=False):
        all_roles_plan = []

        # Store list of roles. Used to compare against DB entities and delete any not found in this list
        operated_on = []

        # Iterate through all entity files, and add the plan to the list of ResourcePlan
        roles_dir = '{}/roles'.format(self.file_path) if not is_org else '{}/roles/orgs/1'.format(self.file_path)
        for filename in os.listdir(roles_dir):
            if filename.endswith(".json") and not self.is_filtered(filter_list, filename[:-5]):
                role_name = filename[:-5]
                role_urn = 'psc:role:{}:{}'.format(self.org_key, role_name) if is_org\
                    else 'psc:role::{}'.format(role_name)
                operated_on.append(role_urn)
                f = open("{}/{}".format(roles_dir, filename), 'r')
                with f:
                    text = f.read()
                    entity_json = json.loads(text)

                    if is_org:
                        try:
                            print("Formatting org role {}".format(entity_json['role_definition']['urn']))
                        except Exception as ex:
                            print("Failed to format json: '{}' with {}".format(entity_json, ex))
                        entity_json = self.format_org_role(entity_json)

                    role_def = entity_json['role_definition']

                    # Create the role plan and default to noop.
                    role_plan = RolePlan(entity_json, "NOOP", "NOOP",
                                             "NOOP", "NOOP")

                    # Get all UUIDs from the local (master) role JSON
                    entity_uuid = role_def['role_uuid'] if 'role_uuid' in role_def else None
                    permissions_uuid = role_def[
                        'role_permission_uuid'] if 'role_permission_uuid' in role_def else None
                    subroles_uuid = role_def['role_subrole_uuid'] if 'role_subrole_uuid' in role_def else None
                    overrides_uuid = role_def['permission_overrides_uuid'] \
                        if 'permission_overrides_uuid' in role_def else None

                    # Determine if the entity needs to be added or deleted
                    api = '/access/v2/roles/{}'.format(role_urn) if not is_org \
                        else '/access/v2/orgs/{}/roles/{}'.format(self.org_key, role_urn)
                    existing_resp = self.session.get(api)

                    if existing_resp.ok:
                        # Retrieve all migration UUIDs from the API response
                        existing_role = json.loads(existing_resp.text)
                        existing_entity_uuid = existing_role[
                            'role_uuid'] if 'role_uuid' in existing_role else None
                        existing_permissions_uuid = existing_role['role_permission_uuid'] \
                            if 'role_permission_uuid' in existing_role else None
                        existing_subroles_uuid = existing_role['role_subrole_uuid'] \
                            if 'role_subrole_uuid' in existing_role else None
                        existing_overrides_uuid = existing_role['permission_overrides_uuid'] \
                            if 'permission_overrides_uuid' in existing_role else None

                        #  If the UUIDs differ, update
                        if existing_entity_uuid != entity_uuid:
                            role_plan.entity_action = "UPDATE"
                        if existing_permissions_uuid != permissions_uuid:
                            role_plan.permissions_action = "UPDATE"
                        if existing_subroles_uuid != subroles_uuid:
                            role_plan.subroles_action = "UPDATE"
                        if existing_overrides_uuid != overrides_uuid:
                            role_plan.overrides_action = "UPDATE"
                    elif existing_resp.status_code == 404:
                        # If the resource does not exist it must be created
                        role_plan.entity_action = "CREATE"
                        role_plan.permissions_action = "CREATE"
                        role_plan.subroles_action = "CREATE"
                        role_plan.overrides_action = "CREATE"
                    else:
                        raise RuntimeError('ERROR: Unexpected response when retrieving a entity definition. '
                                           'Got {}:{}'.format(existing_resp.status_code, existing_resp.text))

                    # If a role's relationships are being updated, the entity must be updated as well to
                    # migrate uuids
                    if (role_plan.permissions_action != "NOOP" \
                        or role_plan.subroles_action != "NOOP" \
                        or role_plan.overrides_action != "NOOP") \
                            and role_plan.entity_action == "NOOP":
                        role_plan.entity_action = "UPDATE"
                    all_roles_plan.append(role_plan)

        # Get all roles via API.  If there are any that no longer exist they should be deleted
        api = '/access/v2/roles' if not is_org else '/access/v2/orgs/{}/roles'.format(self.org_key)
        all_roles_resp = self.session.get(api)
        if not all_roles_resp.ok:
            raise Exception('Problem retrieving all entities from API {}. Got status code: {}. {}'
                            .format(api, all_roles_resp.status_code, all_roles_resp.text))
        all_roles_resp_json = json.loads(all_roles_resp.text)
        for role in all_roles_resp_json:
            urn = role['urn']
            if urn not in operated_on and not self.is_filtered(filter_list, urn):
                role_plan = RolePlan(entity_json={"role_definition": role},
                                         entity_action="DELETE")
                all_roles_plan.append(role_plan)

        return all_roles_plan

    def plan_subrole_migration(self, filter_list):
        all_subroles_plan = []

        # Store list of subroles. Used to compare against DB entities and delete any not found in this list
        operated_on = []

        # Iterate through all entity files, and add the plan to the list of ResourcePlan
        subroles_dir = '{}/subroles'.format(self.file_path)
        for filename in os.listdir(subroles_dir):
            if filename.endswith(".json") and not self.is_filtered(filter_list, filename[:-5]):
                sub_role_id = filename[:-5]
                subrole_urn = 'psc:subrole:{}'.format(sub_role_id)
                operated_on.append(subrole_urn)

                f = open("{}/{}".format(subroles_dir, filename), 'r')
                with f:
                    text = f.read()
                    entity_json = json.loads(text)

                    # Create the subrole plan and default to noop.
                    subrole_plan = SubrolePlan(entity_json, entity_action="NOOP",
                                               permissions_action="NOOP")

                    # Get all UUIDs from the local (master) subrole JSON
                    entity_uuid = entity_json['subrole_uuid'] if 'subrole_uuid' in entity_json else None
                    permissions_uuid = entity_json[
                        'role_permission_uuid'] if 'role_permission_uuid' in entity_json else None

                    # Determine if the entity needs to be added or deleted
                    api = '/access/v2/subroles/{}'.format(subrole_urn)
                    existing_resp = self.session.get(api)

                    if existing_resp.ok:
                        # Retrieve all migration UUIDs from the API response
                        existing_subrole = json.loads(existing_resp.text)
                        existing_entity_uuid = existing_subrole[
                            'subrole_uuid'] if 'subrole_uuid' in existing_subrole else None
                        existing_permissions_uuid = existing_subrole['role_permission_uuid'] \
                            if 'role_permission_uuid' in existing_subrole else None

                        #  If the UUIDs differ, update
                        if existing_entity_uuid != entity_uuid:
                            subrole_plan.entity_action = "UPDATE"
                        if existing_permissions_uuid != permissions_uuid:
                            subrole_plan.permissions_action = "UPDATE"
                    elif existing_resp.status_code == 404:
                        # If the resource does not exist it must be created
                        subrole_plan.entity_action = "CREATE"
                        subrole_plan.permissions_action = "CREATE"
                    else:
                        # If the resource does not exist it must be created
                        subrole_plan.entity_action = "NOOP"
                        subrole_plan.permissions_action = "NOOP"
                        raise RuntimeError('ERROR: Unexpected response when retrieving a entity definition. '
                                           'Got {}:{}'.format(existing_resp.status_code, existing_resp.text))

                    # If a subrole's relationships are being updated, the entity must be updated as well to
                    # migrate uuids
                    if subrole_plan.permissions_action != "NOOP" \
                            and subrole_plan.entity_action == "NOOP":
                        subrole_plan.entity_action = "UPDATE"

                    all_subroles_plan.append(subrole_plan)

        # Get all roles via API.  If there are any that no longer exist they should be deleted
        api = '/access/v2/subroles/list'
        all_subroles_resp = self.session.get(api)
        if not all_subroles_resp.ok:
            raise Exception('Problem retrieving all entities from API {}. Got status code: {}. {}'
                            .format(api, all_subroles_resp.status_code, all_subroles_resp.text))
        all_subroles_resp_json = json.loads(all_subroles_resp.text)
        for subrole in all_subroles_resp_json['subroles']:
            urn = subrole['urn']
            if urn not in operated_on and not self.is_filtered(filter_list, urn):
                subrole_plan = SubrolePlan(subrole, entity_action="DELETE")
                all_subroles_plan.append(subrole_plan)

        return all_subroles_plan

    def plan_permission_migration(self, filter_list):
        all_permissions_plan = []

        # Store list of permissions. Used to compare against DB entities and delete any not found in this list
        operated_on = []

        # Iterate through all entity files, and add the plan to the list of ResourcePlan
        permissions_dir = '{}/permissions'.format(self.file_path)
        for filename in os.listdir(permissions_dir):
            if filename.endswith(".json") and not self.is_filtered(filter_list, filename[:-5]):
                permission_name = filename[:-5]
                permission_urn = 'psc:permission:{}'.format(permission_name)
                operated_on.append(permission_urn)

                f = open('{}/{}'.format(permissions_dir, filename), 'r')
                with f:
                    text = f.read()
                    entity_json = json.loads(text)

                    # Create the permission plan and default to noop.
                    permission_plan = PermissionPlan(entity_json, entity_action="NOOP")

                    # Get UUID from the local (master) permission JSON
                    entity_uuid = entity_json['permission_uuid'] if 'permission_uuid' in entity_json else None

                    # Determine if the entity needs to be added or deleted
                    api = '/access/v2/permissions/{}'.format(permission_urn)
                    existing_resp = self.session.get(api)

                    if existing_resp.ok:
                        # Retrieve all migration UUIDs from the API response
                        existing_permission = json.loads(existing_resp.text)
                        existing_entity_uuid = existing_permission['permission_uuid'] \
                            if 'permission_uuid' in existing_permission else None

                        #  If the UUIDs differ, update
                        if existing_entity_uuid != entity_uuid:
                            permission_plan.entity_action = "UPDATE"
                    elif existing_resp.status_code == 404:
                        # If the resource does not exist it must be created
                        permission_plan.entity_action = "CREATE"
                    else:
                        raise RuntimeError('ERROR: Unexpected response when retrieving a entity definition. '
                                           'Got {}:{}'.format(existing_resp.status_code, existing_resp.text))
                    all_permissions_plan.append(permission_plan)

        # Get all permissions via API.  If there are any that no longer exist they should be deleted
        api = '/access/v2/permissions'
        all_permissions_resp = self.session.get(api)
        if not all_permissions_resp.ok:
            raise Exception('Problem retrieving all entities from API {}. Got status code: {}. {}'
                            .format(api, all_permissions_resp.status_code, all_permissions_resp.text))
        all_permissions_resp_json = json.loads(all_permissions_resp.text)
        for permission in all_permissions_resp_json:
            urn = permission['urn']
            if urn not in operated_on and not self.is_filtered(filter_list, urn):
                subrole_plan = PermissionPlan(permission, entity_action="DELETE")
                all_permissions_plan.append(subrole_plan)

        return all_permissions_plan

    def plan_permission_migration_csp(self, filter_list):
        all_permissions_plan = []

        # Store list of permissions. Used to compare against CSP permissions and delete any not found in this list
        operated_on = []

        # Iterate through all entity files, and add the plan to the list of ResourcePlan
        permissions_dir = '{}/permissions'.format(self.file_path)
        for filename in os.listdir(permissions_dir):
            if filename.endswith(".json") and not self.is_filtered(filter_list, filename[:-5]):
                f = open('{}/{}'.format(permissions_dir, filename), 'r')
                with f:
                    text = f.read()
                    entity_json = json.loads(text)

                    # If the permission does not have the 'public_api' capability we skip its creation in CSP
                    if 'public_api' not in entity_json['capabilities']:
                        continue

                    for action in entity_json['actions']:
                        # Create the permission plan and default to noop.
                        perm_fmt = "_API.{sort_category}:{resource_type}:{action}"
                        permission_name = perm_fmt.format(sort_category=".".join(entity_json['sort_category'].split(' ')),
                                                          resource_type="".join(entity_json['urn'].split(':')[2:]),
                                                          action=action)

                        permission_display_name = entity_json['name'] + " - " + action
                        permission_plan_entity = { "urn": permission_name,
                                                   "displayName": permission_display_name, 
                                                   "description": entity_json['desc'] }

                        permission_plan = PermissionPlan(permission_plan_entity, entity_action="NOOP")

                        operated_on.append(permission_plan_entity['urn'])

                        # Determine if the entity needs to be added or deleted
                        full_csp_permission_name = "{}:{}".format(self.csp_session.service_ticker, permission_plan_entity['urn'])
                        api = '/csp/gateway/iam-roles-mgmt/api/services/{}/permissions/{}'.format(self.csp_session.service_definition_id,
                                                                                                  full_csp_permission_name)
                        retry = True
                        while retry:
                            existing_resp = self.csp_session.get(api)
                            if existing_resp.status_code == 429:
                                time.sleep(60)
                            else:
                                retry = False

                        if existing_resp.ok:
                            existing_permission = json.loads(existing_resp.text)
                            if existing_permission['displayName'] != permission_display_name or existing_permission['description'] != entity_json['desc']:
                                permission_plan.entity_action = "UPDATE"
                        elif existing_resp.status_code == 404:
                            # If the resource does not exist it must be created
                            permission_plan.entity_action = "CREATE"
                        else:
                            raise RuntimeError('ERROR: Unexpected response when retrieving entity definition. '
                                               'Got {}:{}'.format(existing_resp.status_code, existing_resp.text))
                        all_permissions_plan.append(permission_plan)

        # Get all permissions via the CSP API. If there are any that no longer exist they should be deleted
        api = '/csp/gateway/iam-roles-mgmt/api/services/{}/permissions'.format(self.csp_session.service_definition_id)
        all_permissions = []
        more_pages = True
        while more_pages:
            csp_resp = self.csp_session.get(api)
            if not csp_resp.ok:
                raise Exception('Problem retrieving permissions from CSP API {}. Got status code: {}. {}'
                                .format(api, csp_resp.status_code, csp_resp.text))
            csp_resp_json = json.loads(csp_resp.text)
            all_permissions += csp_resp_json['results']
            if "nextLink" in csp_resp_json:
                api = csp_resp_json['nextLink']
            else:
                more_pages = False

        for permission in all_permissions:
            # Cut off the serviceTicker from the permission name
            perm_id = ":".join(permission['id'].split(':')[1:])
            # Skip the UI ones
            if perm_id.startswith("UI."):
                continue

            if perm_id not in operated_on and not self.is_filtered(filter_list, "{}:{}".format("psc:permission", perm_id.split(':')[:-1])):
                delete_permission_entity = { "urn": perm_id,
                                             "displayName": permission['displayName'],
                                             "description": permission['description'] }
                delete_permission_plan = PermissionPlan(delete_permission_entity, entity_action="DELETE")
                all_permissions_plan.append(delete_permission_plan)

        return all_permissions_plan

    def plan_subrole_migration_csp(self, filter_list):
        """NOTE: Subroles are actually permissions in CSP terms"""
        all_subroles_plan = []

        # Store list of subroles. Used to compare against CSP and delete any not found in this list
        operated_on = []

        # Iterate through all entity files, and add the plan to the list of ResourcePlan
        subroles_dir = '{}/subroles'.format(self.file_path)
        for filename in os.listdir(subroles_dir):
            if filename.endswith(".json") and not self.is_filtered(filter_list, filename[:-5]):
                f = open('{}/{}'.format(subroles_dir, filename), 'r')
                with f:
                    text = f.read()
                    entity_json = json.loads(text)

                    # If the subrole does not have the 'public_api' capability we skip its creation in CSP
                    if 'public_api' not in entity_json['capabilities']:
                        continue


                    subrole_fmt = "UI.{sort_category}:{urn}"

                    if not entity_json['desc']:
                        entity_json['desc'] = "No description"
                    elif len(entity_json['desc']) > 256:
                        entity_json['desc'] = entity_json['desc'][:255]


                    subrole_name = subrole_fmt.format(sort_category=".".join(entity_json['sort_category'].strip().replace(',', '').split(' ')),
                            urn=entity_json['urn'].split(':')[-1])

    
                    subrole_plan_entity = { "urn": subrole_name,
                                            "displayName": entity_json['name'], 
                                            "description": entity_json['desc'] }

                    subrole_plan = SubrolePlan(subrole_plan_entity, entity_action="NOOP")

                    operated_on.append(subrole_plan_entity['urn'])

                    # Determine if the entity needs to be added or deleted
                    full_csp_subrole_name = "{}:{}".format(self.csp_session.service_ticker, subrole_plan_entity['urn'])
                    api = '/csp/gateway/iam-roles-mgmt/api/services/{}/permissions/{}'.format(self.csp_session.service_definition_id,
                                                                                              full_csp_subrole_name)
                    retry = True
                    while retry:
                        existing_resp = self.csp_session.get(api)
                        if existing_resp.status_code == 429:
                            time.sleep(60)
                        else:
                            retry = False

                    if existing_resp.ok:
                        existing_subrole = json.loads(existing_resp.text)
                        if existing_subrole['displayName'] != entity_json['name'] or existing_subrole['description'] != entity_json['desc']:
                            subrole_plan.entity_action = "UPDATE"
                    elif existing_resp.status_code == 404:
                        # If the resource does not exist it must be created
                        subrole_plan.entity_action = "CREATE"
                    else:
                        raise RuntimeError('ERROR: Unexpected response when retrieving entity definition. '
                                           'Got {}:{}'.format(existing_resp.status_code, existing_resp.text))
                    all_subroles_plan.append(subrole_plan)

        # Get all subroles via the CSP API. If there are any that no longer exist they should be deleted
        api = '/csp/gateway/iam-roles-mgmt/api/services/{}/permissions'.format(self.csp_session.service_definition_id)
        all_subroles = []
        more_pages = True
        while more_pages:
            csp_resp = self.csp_session.get(api)
            if not csp_resp.ok:
                raise Exception('Problem retrieving permissions from CSP API {}. Got status code: {}. {}'
                                .format(api, csp_resp.status_code, csp_resp.text))
            csp_resp_json = json.loads(csp_resp.text)
            subroles_perms = [ s for s in csp_resp_json['results'] if "UI." in s['name'] ]
            if subroles_perms:
                all_subroles += subroles_perms
            if "nextLink" in csp_resp_json:
                api = csp_resp_json['nextLink']
            else:
                more_pages = False

        for subrole in all_subroles:
            # Cut off the serviceTicker from the subrole name
            subrole_id = ":".join(subrole['id'].split(':')[1:])
            # TODO: Add a better is_filtered() check. Current assumes that the format ot the subrole is known.
            if subrole_id not in operated_on and not self.is_filtered(filter_list, subrole_id):
                delete_subrole_entity = { "urn": subrole['name'],
                                          "displayName": subrole['displayName'],
                                          "description": subrole['description'] }
                delete_subrole_plan = SubrolePlan(delete_subrole_entity, entity_action="DELETE")
                all_subroles_plan.append(delete_subrole_plan)

        return all_subroles_plan

    def is_filtered(self, filter_list, identifier):
        """Returns true if the urn should be removed / filtered"""
        name = identifier if not identifier.startswith("psc:") else identifier.split(':')[-1]
        if filter_list is None:
            return False
        else:
            return name not in filter_list

    def format_org_role(self, org_role_json):
        """Format all instances of "{org_key}" from a org role json"""
        role_urn = org_role_json['role_definition']['urn'].format(org_key=self.org_key)
        org_role_json['role_definition']['urn'] = role_urn
        org_role_json['role_definition']['child_urn'] = role_urn
        org_role_json['role_definition']['scoped'] \
            = org_role_json['role_definition']['scoped'].format(org_key=self.org_key)
        for role_permission in org_role_json['permissions']:
            role_permission['urn'] = role_urn
        for override in org_role_json['overrides']:
            override['urn'] = role_urn
        return org_role_json

