#!/usr/bin/env python3
'''
Platform connectors migration script.
'''

import argparse
import base64
import datetime
import hashlib
import json
import os
import time
import urllib3

import psycopg2
import psycopg2.extras
from kubernetes import client, config

import rbac_util_encoding

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

ENV = os.getenv("ENV")
IS_K8S = (os.getenv("IS_K8S", 'False') == 'true')
FORCE_RECREATE = (os.getenv("forceRecreate", 'False') == 'true')

MULTIENV_DB = os.getenv("MULTIENV_DB", "")
MULTIENV_DB_AUTH = os.getenv("MULTIENV_DB_AUTH", "")
DB_ENDPOINT = os.getenv("DB_ENDPOINT", "")
DB_USER = os.getenv("DB_USER", "")
DB_PASS = os.getenv("DB_PASS", "")
DB_NAME = os.getenv("DB_NAME", "")

AWS_REGION = os.getenv("AWS_REGION", "us-west-2")
MOSAIC_NAMESPACES = [os.getenv("NS_APPS",
                               f"cb-{ENV}-apps"),
                               os.getenv("NS_INFRA",
                               f"cb-{ENV}-infra")]
CONNECTOR_SALT = os.getenv("CONNECTOR_SALT", "")

print("Starting migration")



def parse_arguments():
    '''
    Parse the CLI arguments
    '''
    parser = argparse.ArgumentParser(description="Migrate service connectors from change control.")
    parser.add_argument('-e', '--environment',
                        default=ENV,
                        dest='env',
                        required=False)
    parser.add_argument("--aws-region",
                        default=AWS_REGION,
                        dest='aws_region',
                        required=False,
                        help="The AWS region to connect")
    parser.add_argument("--salt", default=CONNECTOR_SALT,
                        help="The salt use to encode connectorId and apiKey. \
                              If omitted, the org-key will be used as a salt.",
                        dest='salt',
                        required=False)

    return parser.parse_known_args()[0]


def populate_connector_json(connector_json, org_id, salt):
    '''
    Populates the connector JSON
    '''

    connector_name = connector_json['connector_name']
    
    if ENV == 'localdev':
        connector_json['connectorId'] = f"{connector_name}_ID"
        connector_json['apiKey'] = f"{connector_name}_SECRET"
        connector_json['connectorSecret'] = f"{connector_name}_SECRET"
    else:
        connector_json['connectorId'] = rbac_util_encoding.encode_connector_id(connector_name, salt)
        connector_json['apiKey'] = rbac_util_encoding.encode_api_key(connector_name, salt)
        connector_json['connectorSecret'] = rbac_util_encoding.encode_connector_secret(connector_json['apiKey'])

    connector_json.pop('connector_name')
    connector_json.pop('role_urn')

    connector_json['name'] = connector_name
    connector_json['createTime'] = int(time.time())
    connector_json['lastUpdatedTime'] = connector_json['createTime']
    connector_json['changedBy'] = "platform-scmigration"
    connector_json['sessionId'] = None
    connector_json['connectorType'] = "CUSTOM"
    connector_json['orgId'] = org_id
    connector_json['threatExchangeOrgId'] = None
    connector_json['threatExchangeOrgName'] = ""
    connector_json['authorizedIpAddresses'] = []
    connector_json['stats'] = {
                'numRequests': 0,
                'sessionStartTime': None,
                'lastIpAddress': None,
                'lastReportedTime': None
    }
    connector_json['notificationState'] = None
    connector_json['updateVersion'] = 1
    return connector_json


def get_connector_definition(connector_filename, org_id, org_key, salt):
    '''
    Parse the definition of a single connector
    '''
    try:
        with open(f"connectors/orgs/1/{connector_filename}", 'r', encoding='utf-8') as con_file:
            service_connector_json = populate_connector_json(json.loads(con_file.read()),
                                                             org_id,
                                                             salt)
    except IOError:
        print(f"ERROR: An IO error occured when reading from file roles/{connector_filename}")

    result_connector_definition = { "definition": service_connector_json }

    #print(json.dumps(result_connector['definition'], indent=2, sort_keys=True))
    return result_connector_definition


def process_connector(db_conn, con_def):
    '''
    Creates the given connector in the target database
    '''
    ### Setup
    authorized_ip_addresses = con_def['authorizedIpAddresses']
    if not authorized_ip_addresses:
        authorized_ip_addresses = "[]"

    time_format = "%Y-%M-%d %H:%m:%S.%f"
    current_time = datetime.datetime.now().strftime(time_format)[:-4]
    dt_current_time = datetime.datetime.strptime(current_time, time_format)

    cursor = db_conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor)
    cursor.execute("SET DateStyle='ISO,YMD'")

    ### DB ops
    select_statement = """SELECT * FROM connector_identity WHERE connector_id = %s"""
    select_statement_values = (con_def['connectorId'],)

    select_results = []
    print(select_statement % select_statement_values)
    try:
        cursor.execute(select_statement, select_statement_values)
        select_results = cursor.fetchall()
    except Exception as select_exception:
        print(select_exception)
        print(f"""Unable to perform SELECT.
                  No DB operations will execute for connector '{con_def['name']}'.")""")
        return

    insert_flag = False
    update_flag = False
    if select_results:
        print("Connector exists.")
        row = select_results[0]
        # Modify con_def dict, if connector already exists
        con_def['connectorId'] = row['connector_id']
        con_def['apiKey'] = row['identity_key']
        con_def['connectorSecret'] = row['connector_secret']

        if row['org_id'] != str(con_def['orgId']) or \
                row['identity_key_type'] != 'API_KEY' or \
                row['service_connector'] is False or \
                row['description'] != con_def['description'] or \
                row['authorized_ip_addresses'] != authorized_ip_addresses:
            print("The following columns will be updated:")
            print(f"\t'org_id': {row['org_id']} -> {con_def['orgId']}")
            print(f"\t'identity_key_type': {row['identity_key_type']} -> API_KEY")
            print(f"\t'service_connector': {row['service_connector']} -> True")
            print(f"\t'description': {row['description']} -> {con_def['description']}")
            print(f"\t'authorized_ip_addresses': {row['authorized_ip_addresses']} -> {authorized_ip_addresses}")
            print(f"\t'changed_by': {row['changed_by']} -> platform_scmigration")
            print(f"\t'last_updated_time': {row['last_updated_time']} -> {dt_current_time}")
            update_flag = True
        else:
            print("No updates required")
    else:
        print("Connector does not exist.")
        insert_flag = True

    if insert_flag:
        insert_statement = """INSERT INTO connector_identity
                                (connector_id,
                                org_id,
                                identity_key,
                                connector_secret,
                                identity_key_type,
                                service_connector,
                                session_id,
                                name,
                                description,
                                changed_by,
                                create_time,
                                last_updated_time,
                                connector_type,
                                authorized_ip_addresses)
                              VALUES(%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)"""
        insert_statement_values = (con_def['connectorId'],
                                   con_def['orgId'],
                                   con_def['apiKey'],
                                   con_def['connectorSecret'],
                                   'API_KEY',
                                   True,
                                   con_def['sessionId'],
                                   con_def['name'],
                                   con_def['description'],
                                   con_def['changedBy'],
                                   dt_current_time,
                                   dt_current_time,
                                   con_def['connectorType'],
                                   authorized_ip_addresses)
        #print(insert_statement % insert_statement_values)
        try:
            cursor.execute(insert_statement, insert_statement_values)
            db_conn.commit()
            cursor.close()
            print("Connector created.")
            return
        except psycopg2.IntegrityError:
            print("Record already exists.")
            cursor.execute("ROLLBACK")

    if update_flag:
        update_statement = """UPDATE connector_identity SET
                          org_id = %s,
                          identity_key_type = %s,
                          service_connector = %s,
                          description = %s,
                          changed_by = %s,
                          last_updated_time = %s
                          WHERE connector_id = %s AND identity_key = %s"""
        update_statement_values = (con_def['orgId'],
                                   'API_KEY',
                                   True,
                                   con_def['description'],
                                   'platform-scmigration',
                                   dt_current_time,
                                   con_def['connectorId'],
                                   con_def['apiKey'])
        #print(update_statement % update_statement_values)
        cursor.execute(update_statement, update_statement_values)
        db_conn.commit()
        cursor.close()
        print("Connector updated.")

    if FORCE_RECREATE:
        update_statement = """UPDATE connector_identity SET
                          connector_id = %s,
                          org_id = %s,
                          identity_key = %s,
                          connector_secret = %s,
                          identity_key_type = %s,
                          service_connector = %s,
                          session_id = %s,
                          name = %s,
                          description = %s,
                          changed_by = %s,
                          create_time = %s,
                          last_updated_time = %s,
                          connector_type = %s,
                          authorized_ip_addresses = %s,
                          WHERE connector_id = %s AND identity_key = %s"""
        update_statement_values = (con_def['connectorId'],
                                   con_def['orgId'],
                                   con_def['apiKey'],
                                   con_def['connectorSecret'],
                                   'API_KEY',
                                   True,
                                   con_def['sessionId'],
                                   con_def['name'],
                                   con_def['description'],
                                   con_def['changedBy'],
                                   dt_current_time,
                                   dt_current_time,
                                   con_def['connectorType'],
                                   authorized_ip_addresses,
                                   con_def['connectorId'],
                                   con_def['apiKey'])
        #print(update_statement % update_statement_values)
        cursor.execute(update_statement, update_statement_values)
        db_conn.commit()
        cursor.close()
        print("Connector updated.")


def create_connector_secret(core_api_v1, connector):
    '''
    Accepts a Kubernetes client and a connector definition.
    Looks up if existing secret matches the one provided by the connector.
    Recreates the secret if there is a mismatch.
    '''
    secret_name = ""
    secrets_data = {}
    multienv_secrets_data = {}
    # Multienv connectors won't have the 'definition' key at root level
    if 'definition' not in connector.keys():
        for env in connector:
            if connector[env]:
                con_def = connector[env]['definition']
                # For multienv, add the main environment's connector
                # credentials to be backwards compatible
                if ENV == env:
                    secrets_data.update({'connector-id': con_def["connectorId"],
                                         'api-key': con_def["apiKey"]})

                multienv_secrets_data.update({
                                  env: { 'connector-id': con_def['connectorId'],
                                         'api-key': con_def['apiKey']
                                       }
                                })
                if not secret_name:
                    secret_name = con_def['name'].replace('_', '-').lower()
    else:
        connector = connector['definition']
        secret_name = connector['name'].replace('_', '-').lower()
        secrets_data = {'connector-id': connector["connectorId"], 'api-key': connector["apiKey"]}

    for namespace in MOSAIC_NAMESPACES:
        current_secret_exists = False
        try:
            current_secret = core_api_v1.read_namespaced_secret(secret_name, namespace).data
            current_secret_exists = True
        except client.rest.ApiException:
            current_secret = {}

        recreate_secret = False
        if MULTIENV_DB:
            if 'multienv_connector_json' in current_secret:
                current_secret = json.loads(
                                    base64.b64decode(current_secret['multienv_connector_json'])
                                    .decode('utf-8'))
                current_secret.pop('hub-prod-us', None)
                for env in current_secret:
                    ex_env = current_secret[env]
                    if ((connector[env]['definition']['connectorId'] != ex_env['connector-id']) or
                        (connector[env]['definition']['apiKey'] != ex_env['api-key'])):
                        recreate_secret = True
                        break
            else:
                recreate_secret = True
            secrets_data.update({ 'multienv_connector_json': json.dumps(multienv_secrets_data) })
        else:
            ex_connector_id = ""
            ex_api_key = ""
            if current_secret and 'api-key' in current_secret and 'connector-id' in current_secret:
                ex_connector_id = base64.b64decode(current_secret['connector-id']).decode('utf-8')
                ex_api_key = base64.b64decode(current_secret['api-key']).decode('utf-8')
            else:
                recreate_secret = True

            if ((connector["connectorId"] != ex_connector_id) or
                (connector["apiKey"] != ex_api_key)):
                recreate_secret = True

        if recreate_secret:
            new_secret_name = f"{secret_name}-temp"
            new_secret_labels = { "type": "connector", "temp": "true" }
            if current_secret_exists:
                try:
                    core_api_v1.delete_namespaced_secret(new_secret_name, namespace)
                    print(f"Deleted secret '{new_secret_name}' from namespace '{namespace}'...")
                except client.rest.ApiException:
                    print(f"No temp secret to delete in namespace '{namespace}'")
            else:
                new_secret_name = secret_name
                new_secret_labels = { "type": "connector" }

            # Create the new secret
            body = client.V1Secret(api_version = "v1",
                                   kind = "Secret",
                                   metadata = client.V1ObjectMeta(name=new_secret_name,
                                                                  labels=new_secret_labels),
                                   string_data = secrets_data,
                                   type = "Opaque")

            core_api_v1.create_namespaced_secret(namespace, body=body)
            print(f"Created new secret '{new_secret_name}' in namespace '{namespace}'")
        else:
            print(f"No changes for secret '{secret_name}' in namespace '{namespace}'")


def main():
    '''
    Performs the platform-access connector migration.
    '''
    opts = parse_arguments()

    if IS_K8S:
        # Load the cluster config, which is handled by the library, automatically.
        # It is assumed that there is a ServiceAccount attached to the pod, executing this script.
        config.load_incluster_config()
        core_api_v1 = client.CoreV1Api()

    org_id = 1
    org_key = rbac_util_encoding.encode_org_key(env=opts.env, org_id=org_id)
    salt = org_key
    if opts.salt:
        print("Extra salt added")
        salt = salt + opts.salt

    ## DB setup
    multienv = {}
    if MULTIENV_DB and MULTIENV_DB_AUTH:
        print("Multi-environment setup detected.")
        multienv = json.loads(MULTIENV_DB)
        multienv_auth = json.loads(MULTIENV_DB_AUTH)
        print("Environments as per the DB multienv JSON:")
        print(multienv.keys())
        for env in multienv:
            if env not in multienv_auth.keys():
                print(f"No credentials found for {env}. Skipping connector operations for it.")
                continue
            # Creates the DB client for each environment
            db_conn = psycopg2.connect(host=multienv[env]['clusterEndpoint'],
                                       dbname=multienv[env]['databaseName'],
                                       user=multienv_auth[env]['username'],
                                       password=multienv_auth[env]['password'])
            
            env_orgkey = rbac_util_encoding.encode_org_key(env=env, org_id=org_id)
            env_salt = env_orgkey
            if opts.salt:
                print(f"Extra salt added for {env}")
                env_salt = env_salt + opts.salt

            multienv.update({ env: { "db_conn": db_conn,
                                     "org_id": org_id,
                                     "org_key": env_orgkey,
                                     "salt": env_salt
                                   }
                            })
    else:
        db_conn = psycopg2.connect(host=DB_ENDPOINT,
                                   dbname=DB_NAME,
                                   user=DB_USER,
                                   password=DB_PASS)

    ## Connector setup
    for filename in os.listdir('connectors/orgs/1'):
        if filename.endswith(".json"):
            print("-"*80)
            if MULTIENV_DB:
                connector = {}
                for env in multienv:
                    connector.update({ env: get_connector_definition(filename,
                                                                     multienv[env]['org_id'],
                                                                     multienv[env]['org_key'],
                                                                     multienv[env]['salt'])
                    })
            else:
                connector = get_connector_definition(filename, org_id, org_key, salt)
        else:
            continue

        ### Connector operations
        if MULTIENV_DB:
            for env in multienv:
                print(f"Environment: {env}")
                if connector[env]:
                    process_connector(multienv[env]['db_conn'],
                                      connector[env]['definition'])
        else:
            if connector:
                process_connector(db_conn, connector['definition'])

        ### Kubernetes operations
        if IS_K8S:
            create_connector_secret(core_api_v1, connector)

    print("Done migrating service connectors")


if __name__ == "__main__":
    main()
